
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "ril.h"

typedef unsigned long	uLong;
typedef unsigned short	uShort;

static char	*name;

static void	Error(msg)
char		*msg;
{
	fprintf(stderr,"%s : %s\n",name,msg); exit(1);
}

static FILE	*tiff;
static uLong	getBytes;

static void	InitNICDecode(ofs,cnt)
uLong		ofs,cnt;
{
	fseek(tiff,ofs,0); getBytes=cnt;
}

static void	TermNICDecode()
{
	if (getBytes) Error("too many codes.");
}

static Cell	NICDecode()
{
	if (!getBytes) Error("too few codes.");

	--getBytes; return (Cell)fgetc(tiff);
}

static int	numBits,limCode,curBits;
static long	curData;

static int	LZWGet()
{
	while (curBits<numBits) {
	    if (!getBytes) Error("LZW too few codes.");

	    curBits+=8; curData=(curData<<8)|(long)fgetc(tiff); --getBytes;
	}
	curBits-=numBits; return (int)((curData>>curBits)&(long)limCode);
}

#define MINBITS	9
#define MAXBITS	12
#define CLRCODE	(1<<(MINBITS-1))
#define EODCODE	(CLRCODE+1)
#define MINCODE	(CLRCODE+2)
#define MAXCODE	(1<<MAXBITS)
#define STRLEN	(MAXCODE+1)

static int	entCode,strLen;

static void	InitLZWDecode(ofs,cnt)
uLong		ofs,cnt;
{
	fseek(tiff,ofs,0); getBytes=cnt;

	numBits=MINBITS; limCode=(1<<MINBITS)-1; curBits=0;
	if (LZWGet()!=CLRCODE) Error("LZW missing first clear code.");

	curData|=((long)CLRCODE<<curBits); curBits+=numBits;
	entCode=MINCODE; strLen=0;
}

static void	TermLZWDecode()
{
	int	newCode;

	if (entCode==(1<<(numBits-1))-1) {
	    --numBits; limCode=entCode;
	}
	if ((newCode=LZWGet())==CLRCODE) {
	    numBits=MINBITS; limCode=(1<<MINBITS)-1; newCode=LZWGet();
	}
	if (newCode!=EODCODE) Error("LZW missing end of codes.");

	if (strLen || getBytes) Error("LZW too many codes.");
}

static int	oldCode,strBuf[STRLEN+1],head[MAXCODE],tail[MAXCODE];

static Cell	LZWDecode()
{
	int	newCode,idx;
	char	*uec="LZW unexpected end of code.",
		*tls="LZW too long string.";

	if (strLen>0) return (Cell)strBuf[--strLen];

	if ((newCode=LZWGet())==EODCODE) Error(uec);

	if (newCode==CLRCODE) {
	    numBits=MINBITS; limCode=(1<<MINBITS)-1; entCode=MINCODE;

	    if ((newCode=LZWGet())<CLRCODE) {
		*strBuf=newCode; strLen=1;
	    }
	    else
	    if (newCode==EODCODE)
		Error(uec);
	    else
		Error("LZW unknown code after clear code.");
	}
	else {
	    if (newCode<entCode) {
		strLen=0;
		for (idx=newCode; idx>=MINCODE; idx=head[idx]) {
		    strBuf[strLen]=tail[idx]; if (++strLen==STRLEN) Error(tls);
		}
		strBuf[strLen++]=idx;
	    }
	    else
	    if (newCode==entCode) {
		strLen=1;
		for (idx=oldCode; idx>=MINCODE; idx=head[idx]) {
		    strBuf[strLen]=tail[idx]; if (++strLen==STRLEN) Error(tls);
		}
		*strBuf=strBuf[strLen++]=idx;
	    }
	    else
		Error("LZW too large code.");

	    if (entCode==MAXCODE) Error("LZW code table overflow.");

	    head[entCode]=oldCode; tail[entCode]=strBuf[strLen-1];

	    if (++entCode==limCode && numBits<MAXBITS) {
		++numBits; limCode=(1<<numBits)-1;
	    }
	}
	oldCode=newCode; return (Cell)strBuf[--strLen];
}

static int	MPBGet()
{
	if (!getBytes) Error("MPB too few codes.");

	--getBytes; return fgetc(tiff);
}

static int	runLen;

static void	InitMPBDecode(ofs,cnt)
uLong		ofs,cnt;
{
	fseek(tiff,ofs,0); getBytes=cnt;

	runLen=0;
}

static void	TermMPBDecode()
{
	if (runLen || getBytes) Error("MPB too many codes.");
}

static int	repCode;

static Cell	MPBDecode()
{
	if (runLen==0) {
	    while ((runLen=MPBGet())==128) ;

	    if (runLen>128) {
		runLen-=256; return (Cell)(repCode=MPBGet());
	    }
	}
	else
	    if (runLen<0) {
		++runLen; return (Cell)repCode;
	    }
	    else
		--runLen;

	return (Cell)MPBGet();
}

static uShort	Get2I()
{
	uShort	lo=(uShort)fgetc(tiff);

	return ((uShort)fgetc(tiff)<<8)|lo;
}

static uLong	Get4I()
{
	uLong	lo=(uLong)Get2I();

	return ((uLong)Get2I()<<16)|lo;
}

static uShort	Get2M()
{
	uShort	hi=(uShort)fgetc(tiff);

	return (hi<<8)|(uShort)fgetc(tiff);
}

static uLong	Get4M()
{
	uLong	hi=(uLong)Get2M();

	return (hi<<16)|(uLong)Get2M();
}

static uShort	(*Get2)();
static uLong	(*Get4)();

static uShort	GetIFD(ifd,tag)
uLong		ifd;
uShort		tag;
{
	uShort	cnt;

	fseek(tiff,ifd,0);
	for (cnt=(*Get2)(); cnt; cnt--) {
	    if ((*Get2)()==tag) return (*Get2)();

	    fseek(tiff,10L,1);
	}
	return 0;
}

static uLong	GetIFD4(ifd,tag,def)
uLong		ifd,def;
uShort		tag;
{
	switch(GetIFD(ifd,tag)) {
	    case 3 : if ((*Get4)()!=1L) break; return (uLong)(*Get2)();
	    case 4 : if ((*Get4)()!=1L) break; return (*Get4)();
	}
	return def;
}

static uShort	GetIFD3(ifd,tag,def)
uLong		ifd;
uShort		tag,def;
{
	return (GetIFD(ifd,tag)==3 && (*Get4)()==1L)?(*Get2)():def;
}

static int	GetIFDStrip(ifd,tag,spi,sfy,val)
uLong		ifd,spi,sfy,*val;
uShort		tag;
{
	switch(GetIFD(ifd,tag)) {
	    case 3 :
		if ((*Get4)()!=spi) return 2;

		if (spi>2L) fseek(tiff,(uLong)(*Get2)(),0);
		fseek(tiff,sfy*2L,1);
		*val=(uLong)(*Get2)();
		break;
	    case 4 :
		if ((*Get4)()!=spi) return 2;

		if (spi>1L) fseek(tiff,(*Get4)(),0);
		fseek(tiff,sfy*4L,1);
		*val=(*Get4)();
		break;
	    default:
		return 1;
	}
	return 0;
}

static Cell	(*Decode)();

static Cell	BitReverse()
{
	Cell	data=(*Decode)();

	return ((data&1)<<7)|((data&2)<<5)|((data&4)<<3)|((data&8)<<1)|
	       ((data>>1)&8)|((data>>3)&4)|((data>>5)&2)|((data>>7)&1);
}

#define MA(cnt,ptr)	malloc((size_t)(cnt)*sizeof(*(ptr)))

void	ReadImageLine(path,Nx,Ny,BPS,cell,y)
char	*path;
int	*Nx,*Ny,*BPS,y;
Cell	**cell;
{
	int	code,swap,x,idx,bit;
	uLong	ifd,iw,il,rps,spi,sfy,ofs,cnt;
	uShort	pi,bps;
	void	(*InitDecode)(),(*TermDecode)();
	Cell	(*GetData)(),mask,*line,data;

	name=path;

	if ((tiff=fopen(path,"rb"))==NULL) Error("file not found.");

	switch(code=fgetc(tiff)) {
	    case 'I' : Get4=Get4I; Get2=Get2I; swap=1; break;
	    case 'M' : Get4=Get4M; Get2=Get2M; swap=0; break;
	    default  : code=0; break;
	}
	if (!code || fgetc(tiff)!=code || (*Get2)()!=0x002a)
	    Error("bad magic number.");

	ifd=(*Get4)();

	if ((iw=GetIFD4(ifd,0x100,0L))==0L ||
	    (il=GetIFD4(ifd,0x101,0L))==0L ||
	    (pi=GetIFD3(ifd,0x106,0xffff))==0xffff)
	    Error("primary image attribute not found.");

	if (iw>>(8*sizeof(int)-1) ||
	    il>>(8*sizeof(int)-1)) Error("too large image size.");

	if (pi!=0 && pi!=1 && pi!=3 && pi!=4)
	    Error("bad photometric interpretation.");

	if (GetIFD3(ifd,0x115,1)>1) Error("too many samples per pixel.");

	if ((bps=GetIFD3(ifd,0x102,1))>8*sizeof(Cell))
	    Error("too many bits per sample.");

	if (GetIFD3(ifd,0x11c,1)!=1) Error("bad planar configuration.");

	switch(GetIFD3(ifd,0x103,1)) {
	    case 1     : InitDecode=InitNICDecode;
		         TermDecode=TermNICDecode; Decode=NICDecode; break;
	    case 5     : InitDecode=InitLZWDecode;
		         TermDecode=TermLZWDecode; Decode=LZWDecode; break;
	    case 32773 : InitDecode=InitMPBDecode;
		         TermDecode=TermMPBDecode; Decode=MPBDecode; break;
	    default    : Error("unknown image compression method.");
	}
	switch(GetIFD3(ifd,0x10a,1)) {
	    case 1 : GetData=Decode; break;
	    case 2 : GetData=BitReverse; break;
	    default: Error("bad fill order.");
	}
	switch(GetIFD3(ifd,0x13d,1)) {
	    case 1 : mask=0; break;
	    case 2 : mask=(Cell)((1<<bps)-1); break;
	    default: Error("unknown differencing predictor.");
	}

	if (Nx!=NULL) *Nx=(int)iw;
	if (Ny!=NULL) *Ny=(int)il; if (BPS!=NULL) *BPS=(int)bps;

	if (cell!=NULL)
{
	if (y<0 || y>=(int)il) Error("bad line position.");

	if ((*cell=line=(Cell *)MA(iw,*cell))==NULL)
	    Error("cell memory allocation error.");

	rps=GetIFD4(ifd,0x116,il); spi=(il+rps-1L)/rps; sfy=(long)y/rps;

	switch(GetIFDStrip(ifd,0x111,spi,sfy,&ofs)) {
	    case 1 : Error("strip offset not found.");
	    case 2 : Error("bad number of strip offset.");
	}
	switch(GetIFDStrip(ifd,0x117,spi,sfy,&cnt)) {
	    case 1 : Error("strip bytes count not found.");
	    case 2 : Error("bad number of strip bytes count.");
	}

	(*InitDecode)(ofs,cnt);

	for (idx=(y%(int)rps)*(((int)iw*(int)bps+7)/8); idx>0; idx--) 
	    (void)(*Decode)();

	if (bps==8)
	    for (x=0; x<(int)iw; x++) line[x]=(*GetData)();
	else if (bps==16)
	    if (swap)
		for (x=0; x<(int)iw; x++) line[x]= (*GetData)()    |
						  ((*GetData)()<<8);
	    else
		for (x=0; x<(int)iw; x++) line[x]=((*GetData)()<<8)|
						   (*GetData)()    ;
	else if (bps==32)
	    if (swap)
		for (x=0; x<(int)iw; x++) line[x]= (*GetData)()     |
						  ((*GetData)()<< 8)|
						  ((*GetData)()<<16)|
						  ((*GetData)()<<24);
	    else
		for (x=0; x<(int)iw; x++) line[x]=((*GetData)()<<24)|
						  ((*GetData)()<<16)|
						  ((*GetData)()<< 8)|
						   (*GetData)()     ;
	else
	    for (bit=0x80, x=0; x<(int)iw; x++) {
		for (data=0, idx=0; idx<(int)bps; idx++) {
		    if (bit==0x80) code=(*GetData)();
		    data<<=1; if (code&bit) data|=1;
		    if (bit==0x01) bit=0x80; else bit>>=1;
		}
		line[x]=data;
	    }

	if (y+1==(int)il || (y+1)%(int)rps==0) (*TermDecode)();

	if (mask) {
	    data=line[0];
	    for (x=1; x<(int)iw; x++) {
		data=(data+line[x])&mask; line[x]=data;
	    }
	}
}
	fclose(tiff);
}
